@@ -12,6 +12,10 @@
 # License for the specific language governing permissions and limitations under
 # the License.
 
+# Python3 support
+from __future__ import print_function
+from __future__ import unicode_literals
+
 # std libs
 import sys
 
@@ -19,14 +23,19 @@ import sys
 import napalm_base.exceptions
 import napalm_base.helpers
 
+import napalm_base.constants as c
+
+from napalm_base import validate
+
 
 class NetworkDriver(object):
 
-    def __init__(self, hostname, username, password, timeout, optional_args):
+    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
         """
-        This is the base class you have to inherit from when writing your own Network Driver to manage any device. You
-        will, in addition, have to override all the methods specified on this class. Make sure you follow the guidelines
-        for every method and that you return the correct data.
+        This is the base class you have to inherit from when writing your own Network Driver to
+        manage any device. You will, in addition, have to override all the methods specified on
+        this class. Make sure you follow the guidelines for every method and that you return the
+        correct data.
 
         :param hostname: (str) IP or FQDN of the device you want to connect to.
         :param username: (str) Username you want to use
@@ -53,21 +62,24 @@ class NetworkDriver(object):
     @staticmethod
     def __raise_clean_exception(exc_type, exc_value, exc_traceback):
         """
-        This method is going to check if the exception exc_type is part of the builtins exceptions or part of the
-        napalm exceptions. If it is not, it will print a message on the screen giving instructions to fill a bug.
+        This method is going to check if the exception exc_type is part of the builtins exceptions
+        or part of the napalm exceptions. If it is not, it will print a message on the screen
+        giving instructions to fill a bug.
+
         Finally it will raise the original exception.
 
         :param exc_type: Exception class.
         :param exc_value: Exception object.
         :param exc_traceback: Traceback.
         """
-        if exc_type.__name__ not in dir(napalm_base.exceptions) and \
-                        exc_type.__name__ not in __builtins__.keys():
+        if (exc_type.__name__ not in dir(napalm_base.exceptions) and
+                exc_type.__name__ not in __builtins__.keys()):
             epilog = ("NAPALM didn't catch this exception. Please, fill a bugfix on "
                       "https://github.com/napalm-automation/napalm/issues\n"
                       "Don't forget to include this traceback.")
             print(epilog)
-        raise exc_type, exc_value, exc_traceback
+        # Traceback should already be attached to exception; no need to re-attach
+        raise exc_value
 
     def open(self):
         """
@@ -81,21 +93,32 @@ class NetworkDriver(object):
         """
         raise NotImplementedError
 
-    def load_template(self, template_name, template_source=None, template_path=None, **template_vars):
+    def is_alive(self):
+        """
+        Returns a flag with the connection state.
+        Depends on the nature of API used by each driver.
+        The state does not reflect only on the connection status (when SSH), it must also take into
+        consideration other parameters, e.g.: NETCONF session might not be usable, althought the
+        underlying SSH session is still open etc.
+        """
+        raise NotImplementedError
+
+    def load_template(self, template_name, template_source=None,
+                      template_path=None, **template_vars):
         """
         Will load a templated configuration on the device.
 
         :param cls: Instance of the driver class.
         :param template_name: Identifies the template name.
-        :param template_source (optional): A custom config template to be rendered and loaded on the device
-        :param template_path (optional): Specifies the absolute path to a different directory for the configuration \
-        templates
-        :param template_vars: Dictionary with the arguments  to be used when the template is rendered.
-        :raise DriverTemplateNotImplemented: No template defined for the device type
-        :raise TemplateNotImplemented: The template specified in template_name does not exist in the default path or \
-        in the custom path if any specified using parameter `template_path`
-        :raise TemplateRenderException: The template could not be rendered. Either the template source does not have \
-        the right format, either the arguments in `template_vars` are not properly specified.
+        :param template_source (optional): Custom config template rendered and loaded on device
+        :param template_path (optional): Absolute path to directory for the configuration templates
+        :param template_vars: Dictionary with arguments to be used when the template is rendered.
+        :raise DriverTemplateNotImplemented: No template defined for the device type.
+        :raise TemplateNotImplemented: The template specified in template_name does not exist in \
+        the default path or in the custom path if any specified using parameter `template_path`.
+        :raise TemplateRenderException: The template could not be rendered. Either the template \
+        source does not have the right format, either the arguments in `template_vars` are not \
+        properly specified.
         """
         return napalm_base.helpers.load_template(self,
                                                  template_name,
@@ -105,11 +128,13 @@ class NetworkDriver(object):
 
     def load_replace_candidate(self, filename=None, config=None):
         """
-        Populates the candidate configuration. You can populate it from a file or from a string. If you send both a
-        filename and a string containing the configuration, the file takes precedence.
+        Populates the candidate configuration. You can populate it from a file or from a string.
+        If you send both a filename and a string containing the configuration, the file takes
+        precedence.
 
-        If you use this method the existing configuration will be replaced entirely by the candidate configuration once
-        you commit the changes. This method will not change the configuration by itself.
+        If you use this method the existing configuration will be replaced entirely by the
+        candidate configuration once you commit the changes. This method will not change the
+        configuration by itself.
 
         :param filename: Path to the file containing the desired configuration. By default is None.
         :param config: String containing the desired configuration.
@@ -119,11 +144,13 @@ class NetworkDriver(object):
 
     def load_merge_candidate(self, filename=None, config=None):
         """
-        Populates the candidate configuration. You can populate it from a file or from a string. If you send both a
-        filename and a string containing the configuration, the file takes precedence.
+        Populates the candidate configuration. You can populate it from a file or from a string.
+        If you send both a filename and a string containing the configuration, the file takes
+        precedence.
 
-        If you use this method the existing configuration will be merged with the candidate configuration once
-        you commit the changes. This method will not change the configuration by itself.
+        If you use this method the existing configuration will be merged with the candidate
+        configuration once you commit the changes. This method will not change the configuration
+        by itself.
 
         :param filename: Path to the file containing the desired configuration. By default is None.
         :param config: String containing the desired configuration.
@@ -133,8 +160,9 @@ class NetworkDriver(object):
 
     def compare_config(self):
         """
-        :return: A string showing the difference between the running configuration and the candidate configuration. The\
-        running_config is loaded automatically just before doing the comparison so there is no need for you to do it.
+        :return: A string showing the difference between the running configuration and the \
+        candidate configuration. The running_config is loaded automatically just before doing the \
+        comparison so there is no need for you to do it.
         """
         raise NotImplementedError
 
@@ -186,8 +214,9 @@ class NetworkDriver(object):
 
     def get_interfaces(self):
         """
-        Returns a dictionary of dictionaries. The keys for the first dictionary will be the interfaces in the devices.\
-        The inner dictionary will containing the following data for each interface:
+        Returns a dictionary of dictionaries. The keys for the first dictionary will be the \
+        interfaces in the devices. The inner dictionary will containing the following data for \
+        each interface:
          * is_up (True/False)
          * is_enabled (True/False)
          * description (string)
@@ -240,8 +269,8 @@ class NetworkDriver(object):
 
     def get_lldp_neighbors(self):
         """
-        Returns a dictionary where the keys are local ports and the value is a list of dictionaries with the following \
-        information:
+        Returns a dictionary where the keys are local ports and the value is a list of \
+        dictionaries with the following information:
             * hostname
             * port
 
@@ -286,11 +315,12 @@ class NetworkDriver(object):
 
     def get_bgp_neighbors(self):
         """
-        Returns a dictionary of dictionaries. The keys for the first dictionary will be the vrf (global if no vrf).
-        The inner dictionary will contain the following data for each vrf:
+        Returns a dictionary of dictionaries. The keys for the first dictionary will be the vrf
+        (global if no vrf). The inner dictionary will contain the following data for each vrf:
 
           * router_id
-          * peers - another dictionary of dictionaries. Outer keys are the IPs of the neighbors. The inner keys are:
+          * peers - another dictionary of dictionaries. Outer keys are the IPs of the neighbors. \
+            The inner keys are:
              * local_as (int)
              * remote_as (int)
              * remote_id - peer router id
@@ -298,8 +328,8 @@ class NetworkDriver(object):
              * is_enabled (True/False)
              * description (string)
              * uptime (int in seconds)
-             * address_family (dictionary) - A dictionary of address families available for the neighbor. So far it can\
-               be 'ipv4' or 'ipv6'
+             * address_family (dictionary) - A dictionary of address families available for the \
+               neighbor. So far it can be 'ipv4' or 'ipv6'
                 * received_prefixes (int)
                 * accepted_prefixes (int)
                 * sent_prefixes (int)
@@ -312,10 +342,10 @@ class NetworkDriver(object):
 
             * fans is a dictionary of dictionaries where the key is the location and the values:
                  * status (True/False) - True if it's ok, false if it's broken
-            * temperature is a dictionary of dictionaries where the key is the location and the values:
+            * temperature is a dict of dictionaries where the key is the location and the values:
                  * temperature (float) - Temperature in celsius the sensor is reporting.
                  * is_alert (True/False) - True if the temperature is above the alert threshold
-                 * is_critical (True/False) - True if the temperature is above the critical threshold
+                 * is_critical (True/False) - True if the temp is above the critical threshold
             * power is a dictionary of dictionaries where the key is the PSU id and the values:
                  * status (True/False) - True if it's ok, false if it's broken
                  * capacity (float) - Capacity in W that the power supply can support
@@ -330,8 +360,8 @@ class NetworkDriver(object):
 
     def get_interfaces_counters(self):
         """
-        Returns a dictionary of dictionaries where the first key is an interface name and the inner dictionary contains
-        the following keys:
+        Returns a dictionary of dictionaries where the first key is an interface name and the
+        inner dictionary contains the following keys:
 
             * tx_errors (int)
             * rx_errors (int)
@@ -395,7 +425,7 @@ class NetworkDriver(object):
         """
         raise NotImplementedError
 
-    def get_lldp_neighbors_detail(self, interface = ''):
+    def get_lldp_neighbors_detail(self, interface=''):
         """
         Returns a detailed view of the LLDP neighbors as a dictionary
         containing lists of dictionaries for each interface.
@@ -420,7 +450,8 @@ class NetworkDriver(object):
                         'remote_system_name': u'switch',
                         'remote_port': u'Eth2/2/1',
                         'remote_port_description': u'Ethernet2/2/1',
-                        'remote_system_description': u'''Cisco Nexus Operating System (NX-OS) Software 7.1(0)N1(1a)
+                        'remote_system_description': u'''Cisco Nexus Operating System (NX-OS)
+                              Software 7.1(0)N1(1a)
                               TAC support: http://www.cisco.com/tac
                               Copyright (c) 2002-2015, Cisco Systems, Inc. All rights reserved.''',
                         'remote_system_capab': u'B, R',
@@ -439,7 +470,8 @@ class NetworkDriver(object):
         :param group: Returns the configuration of a specific BGP group.
         :param neighbor: Returns the configuration of a specific BGP neighbor.
 
-        Main dictionary keys represent the group name and the values represent a dictionary having the following keys:
+        Main dictionary keys represent the group name and the values represent a dictionary having
+        the following keys:
             * type (string)
             * description (string)
             * apply_groups (string list)
@@ -525,7 +557,7 @@ class NetworkDriver(object):
         """
         raise NotImplementedError
 
-    def cli(self, *commands):
+    def cli(self, commands):
 
         """
         Will execute a list of commands and return the output in a dictionary format.
@@ -541,14 +573,14 @@ class NetworkDriver(object):
                                                         I just saw Episode Two
                                                         You're my only hope
                                             ''',
-                u'show chassis fan'     :   u'''Item                      Status   RPM     Measurement
-                                                Top Rear Fan              OK       3840    Spinning at intermediate-speed
-                                                Bottom Rear Fan           OK       3840    Spinning at intermediate-speed
-                                                Top Middle Fan            OK       3900    Spinning at intermediate-speed
-                                                Bottom Middle Fan         OK       3840    Spinning at intermediate-speed
-                                                Top Front Fan             OK       3810    Spinning at intermediate-speed
-                                                Bottom Front Fan          OK       3840    Spinning at intermediate-speed
-                                            '''
+                u'show chassis fan'     :   u'''
+                    Item               Status  RPM     Measurement
+                    Top Rear Fan       OK      3840    Spinning at intermediate-speed
+                    Bottom Rear Fan    OK      3840    Spinning at intermediate-speed
+                    Top Middle Fan     OK      3900    Spinning at intermediate-speed
+                    Bottom Middle Fan  OK      3840    Spinning at intermediate-speed
+                    Top Front Fan      OK      3810    Spinning at intermediate-speed
+                    Bottom Front Fan   OK      3840    Spinning at intermediate-speed'''
             }
         """
         raise NotImplementedError
@@ -560,11 +592,15 @@ class NetworkDriver(object):
 
         :param neighbor_address: Retuns the statistics for a spcific BGP neighbor.
 
-        The keys of the dictionary represent the AS number of the neighbors.
-        Inner dictionaries contain the following fields:
+        Returns a dictionary of dictionaries. The keys for the first dictionary will be the vrf
+        (global if no vrf).
+        The keys of the inner dictionary represent the AS number of the neighbors.
+        Leaf dictionaries contain the following fields:
+
             * up (True/False)
             * local_as (int)
             * remote_as (int)
+            * router_id (string)
             * local_address (string)
             * routing_table (string)
             * local_address_configured (True/False)
@@ -600,44 +636,46 @@ class NetworkDriver(object):
         Example::
 
             {
-                8121: [
-                    {
-                        'up'                        : True,
-                        'local_as'                  : 13335,
-                        'remote_as'                 : 8121,
-                        'local_address'             : u'172.101.76.1',
-                        'local_address_configured'  : True,
-                        'local_port'                : 179,
-                        'routing_table'             : u'inet.0',
-                        'remote_address'            : u'192.247.78.0',
-                        'remote_port'               : 58380,
-                        'multihop'                  : False,
-                        'multipath'                 : True,
-                        'remove_private_as'         : True,
-                        'import_policy'             : u'4-NTT-TRANSIT-IN',
-                        'export_policy'             : u'4-NTT-TRANSIT-OUT',
-                        'input_messages'            : 123,
-                        'output_messages'           : 13,
-                        'input_updates'             : 123,
-                        'output_updates'            : 5,
-                        'messages_queued_out'       : 23,
-                        'connection_state'          : u'Established',
-                        'previous_connection_state' : u'EstabSync',
-                        'last_event'                : u'RecvKeepAlive',
-                        'suppress_4byte_as'         : False,
-                        'local_as_prepend'          : False,
-                        'holdtime'                  : 90,
-                        'configured_holdtime'       : 90,
-                        'keepalive'                 : 30,
-                        'configured_keepalive'      : 30,
-                        'active_prefix_count'       : 132808,
-                        'received_prefix_count'     : 566739,
-                        'accepted_prefix_count'     : 566479,
-                        'suppressed_prefix_count'   : 0,
-                        'advertise_prefix_count'    : 0,
-                        'flap_count'                : 27
-                    }
-                ]
+                'global': {
+                    8121: [
+                        {
+                            'up'                        : True,
+                            'local_as'                  : 13335,
+                            'remote_as'                 : 8121,
+                            'local_address'             : u'172.101.76.1',
+                            'local_address_configured'  : True,
+                            'local_port'                : 179,
+                            'routing_table'             : u'inet.0',
+                            'remote_address'            : u'192.247.78.0',
+                            'remote_port'               : 58380,
+                            'multihop'                  : False,
+                            'multipath'                 : True,
+                            'remove_private_as'         : True,
+                            'import_policy'             : u'4-NTT-TRANSIT-IN',
+                            'export_policy'             : u'4-NTT-TRANSIT-OUT',
+                            'input_messages'            : 123,
+                            'output_messages'           : 13,
+                            'input_updates'             : 123,
+                            'output_updates'            : 5,
+                            'messages_queued_out'       : 23,
+                            'connection_state'          : u'Established',
+                            'previous_connection_state' : u'EstabSync',
+                            'last_event'                : u'RecvKeepAlive',
+                            'suppress_4byte_as'         : False,
+                            'local_as_prepend'          : False,
+                            'holdtime'                  : 90,
+                            'configured_holdtime'       : 90,
+                            'keepalive'                 : 30,
+                            'configured_keepalive'      : 30,
+                            'active_prefix_count'       : 132808,
+                            'received_prefix_count'     : 566739,
+                            'accepted_prefix_count'     : 566479,
+                            'suppressed_prefix_count'   : 0,
+                            'advertised_prefix_count'   : 0,
+                            'flap_count'                : 27
+                        }
+                    ]
+                }
             }
         """
         raise NotImplementedError
@@ -671,7 +709,6 @@ class NetworkDriver(object):
         """
         raise NotImplementedError
 
-
     def get_ntp_peers(self):
 
         """
@@ -692,6 +729,25 @@ class NetworkDriver(object):
 
         raise NotImplementedError
 
+    def get_ntp_servers(self):
+
+        """
+        Returns the NTP servers configuration as dictionary.
+        The keys of the dictionary represent the IP Addresses of the servers.
+        Inner dictionaries do not have yet any available keys.
+
+        Example::
+
+            {
+                '192.168.0.1': {},
+                '17.72.148.53': {},
+                '37.187.56.220': {},
+                '162.158.20.18': {}
+            }
+
+        """
+
+        raise NotImplementedError
 
     def get_ntp_stats(self):
 
@@ -736,7 +792,8 @@ class NetworkDriver(object):
         Returns all configured IP addresses on all interfaces as a dictionary of dictionaries.
         Keys of the main dictionary represent the name of the interface.
         Values of the main dictionary represent are dictionaries that may consist of two keys
-        'ipv4' and 'ipv6' (one, both or none) which are themselvs dictionaries witht the IP addresses as keys.
+        'ipv4' and 'ipv6' (one, both or none) which are themselvs dictionaries witht the IP
+        addresses as keys.
         Each IP Address dictionary has the following keys:
             * prefix_length (int)
 
@@ -785,8 +842,8 @@ class NetworkDriver(object):
     def get_mac_address_table(self):
 
         """
-        Returns a lists of dictionaries. Each dictionary represents an entry in the MAC Address Table,
-        having the following keys:
+        Returns a lists of dictionaries. Each dictionary represents an entry in the MAC Address
+        Table, having the following keys:
             * mac (string)
             * interface (string)
             * vlan (int)
@@ -835,7 +892,8 @@ class NetworkDriver(object):
     def get_route_to(self, destination='', protocol=''):
 
         """
-        Returns a dictionary of dictionaries containing details of all available routes to a destination.
+        Returns a dictionary of dictionaries containing details of all available routes to a
+        destination.
 
         :param destination: The destination prefix to be used when filtering the routes.
         :param protocol (optional): Retrieve the routes only for a specific protocol.
@@ -950,7 +1008,8 @@ class NetworkDriver(object):
     def get_probes_config(self):
         """
         Returns a dictionary with the probes configured on the device.
-        Probes can be either RPM on JunOS devices, either SLA on IOS-XR. Other vendors do not support probes.
+        Probes can be either RPM on JunOS devices, either SLA on IOS-XR. Other vendors do not
+        support probes.
         The keys of the main dictionary represent the name of the probes.
         Each probe consists on multiple tests, each test name being a key in the probe dictionary.
         A test has the following keys:
@@ -1051,8 +1110,8 @@ class NetworkDriver(object):
         """
         raise NotImplementedError
 
-
-    def ping(self, destination, source='', ttl=0, timeout=0, size=0, count=0):
+    def ping(self, destination, source=c.PING_SOURCE, ttl=c.PING_TTL, timeout=c.PING_TIMEOUT,
+             size=c.PING_SIZE, count=c.PING_COUNT):
         """
         Executes ping on the device and returns a dictionary with the result
 
@@ -1115,8 +1174,11 @@ class NetworkDriver(object):
         """
         raise NotImplementedError
 
-
-    def traceroute(self, destination, source='', ttl=0, timeout=0):
+    def traceroute(self,
+                   destination,
+                   source=c.TRACEROUTE_SOURCE,
+                   ttl=c.TRACEROUTE_TTL,
+                   timeout=c.TRACEROUTE_TIMEOUT):
         """
         Executes traceroute on the device and returns a dictionary with the result.
 
@@ -1130,9 +1192,8 @@ class NetworkDriver(object):
             * success
             * error
 
-        In case of success, the keys of the dictionary represent the hop ID, while values are dictionaries
-        containing the probes results:
-
+        In case of success, the keys of the dictionary represent the hop ID, while values are
+        dictionaries containing the probes results:
             * rtt (float)
             * ip_address (str)
             * host_name (str)
@@ -1223,18 +1284,17 @@ class NetworkDriver(object):
             """
         raise NotImplementedError
 
-
     def get_users(self):
         """
         Returns a dictionary with the configured users.
-        The keys of the main dictionary represents the username. The values represent the details of the user,
-        represented by the following keys:
-
+        The keys of the main dictionary represents the username. The values represent the details
+        of the user, represented by the following keys:
             * level (int)
             * password (str)
             * sshkeys (list)
 
-        The level is an integer between 0 and 15, where 0 is the lowest access and 15 represents full access to the device.
+        The level is an integer between 0 and 15, where 0 is the lowest access and 15 represents
+        full access to the device.
 
         Example::
 
@@ -1243,9 +1303,199 @@ class NetworkDriver(object):
                     'level': 15,
                     'password': '$1$0P70xKPa$z46fewjo/10cBTckk6I/w/',
                     'sshkeys': [
-                        'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC4pFn+shPwTb2yELO4L7NtQrKOJXNeCl1jel9STXVaGnRAnuc2PXl35vnWmcUq6YbUEcgUTRzzXfmelJKuVJTJIlMXii7h2xkbQp0YZIEs4P8ipwnRBAxFfk/ZcDsN3mjep4/yjN56eorF5xs7zP9HbqbJ1dsqk1p3A/9LIL7l6YewLBCwJj6D+fWSJ0/YW+7oH17Fk2HH+tw0L5PcWLHkwA4t60iXn16qDbIk/ze6jv2hDGdCdz7oYQeCE55CCHOHMJWYfN3jcL4s0qv8/u6Ka1FVkV7iMmro7ChThoV/5snI4Ljf2wKqgHH7TfNaCfpU0WvHAnTs8zhOrGScSrtb mircea@master-roshi'
+                        'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC4pFn+shPwTb2yELO4L7NtQrKOJXNeCl1je\
+                         l9STXVaGnRAnuc2PXl35vnWmcUq6YbUEcgUTRzzXfmelJKuVJTJIlMXii7h2xkbQp0YZIEs4P\
+                         8ipwnRBAxFfk/ZcDsN3mjep4/yjN56eorF5xs7zP9HbqbJ1dsqk1p3A/9LIL7l6YewLBCwJj6\
+                         D+fWSJ0/YW+7oH17Fk2HH+tw0L5PcWLHkwA4t60iXn16qDbIk/ze6jv2hDGdCdz7oYQeCE55C\
+                         CHOHMJWYfN3jcL4s0qv8/u6Ka1FVkV7iMmro7ChThoV/5snI4Ljf2wKqgHH7TfNaCfpU0WvHA\
+                         nTs8zhOrGScSrtb mircea@master-roshi'
                     ]
                 }
             }
         """
         raise NotImplementedError
+
+    def get_optics(self):
+        """Fetches the power usage on the various transceivers installed
+        on the switch (in dbm), and returns a view that conforms with the
+        openconfig model openconfig-platform-transceiver.yang
+
+        Returns a dictionary where the keys are as listed below:
+
+            * intf_name (unicode)
+                * physical_channels
+                    * channels (list of dicts)
+                        * index (int)
+                        * state
+                            * input_power
+                                * instant (float)
+                                * avg (float)
+                                * min (float)
+                                * max (float)
+                            * output_power
+                                * instant (float)
+                                * avg (float)
+                                * min (float)
+                                * max (float)
+                            * laser_bias_current
+                                * instant (float)
+                                * avg (float)
+                                * min (float)
+                                * max (float)
+
+        Example:
+
+            {
+                    'et1': {
+                        'physical_channels': {
+                            'channel': [
+                                {
+                                    'index': 0,
+                                    'state': {
+                                        'input_power': {
+                                            'instant': 0.0,
+                                            'avg': 0.0,
+                                            'min': 0.0,
+                                            'max': 0.0,
+                                        },
+                                        'output_power': {
+                                            'instant': 0.0,
+                                            'avg': 0.0,
+                                            'min': 0.0,
+                                            'max': 0.0,
+                                        },
+                                        'laser_bias_current': {
+                                            'instant': 0.0,
+                                            'avg': 0.0,
+                                            'min': 0.0,
+                                            'max': 0.0,
+                                        },
+                                    }
+                                }
+                            ]
+                        }
+                    }
+                }
+        """
+        raise NotImplementedError
+
+    def get_config(self, retrieve='all'):
+        """
+        Return the configuration of a device.
+
+        Args:
+            retrieve(string): Which configuration type you want to populate, default is all of them.
+                The rest will be set to "".
+
+        Returns:
+          The object returned is a dictionary with the following keys:
+            - running(string) - Representation of the native running configuration
+            - candidate(string) - Representation of the native candidate configuration. If the
+              device doesnt differentiate between running and startup configuration this will an
+              empty string
+            - startup(string) - Representation of the native startup configuration. If the
+              device doesnt differentiate between running and startup configuration this will an
+              empty string
+        """
+        raise NotImplementedError
+
+    def get_network_instances(self, name=''):
+        """
+        Return a dictionary of network instances (VRFs) configured, including default/global
+
+        Args:
+            name(string) - Name of the network instance to return, default is all.
+
+        Returns:
+            A dictionary of network instances in OC format:
+            * name (dict)
+              * name (unicode)
+              * type (unicode)
+              * state (dict)
+                * route_distinguisher (unicode)
+              * interfaces (dict)
+                * interface (dict)
+                  * interface name: (dict)
+
+        Example:
+        {
+            u'MGMT': {
+                u'name': u'MGMT',
+                u'type': u'L3VRF',
+                u'state': {
+                    u'route_distinguisher': u'123:456',
+                },
+                u'interfaces': {
+                    u'interface': {
+                        u'Management1': {}
+                    }
+                }
+            }
+            u'default': {
+                u'name': u'default',
+                u'type': u'DEFAULT_INSTANCE',
+                u'state': {
+                    u'route_distinguisher': None,
+                },
+                u'interfaces: {
+                    u'interface': {
+                        u'Ethernet1': {}
+                        u'Ethernet2': {}
+                        u'Ethernet3': {}
+                        u'Ethernet4': {}
+                    }
+                }
+            }
+        }
+        """
+        raise NotImplementedError
+
+    def get_firewall_policies(self):
+        """
+        Returns a dictionary of lists of dictionaries where the first key is an unique policy
+        name and the inner dictionary contains the following keys:
+
+        * position (int)
+        * packet_hits (int)
+        * byte_hits (int)
+        * id (text_type)
+        * enabled (bool)
+        * schedule (text_type)
+        * log (text_type)
+        * l3_src (text_type)
+        * l3_dst (text_type)
+        * service (text_type)
+        * src_zone (text_type)
+        * dst_zone (text_type)
+        * action (text_type)
+
+        Example::
+
+        {
+            'policy_name': [{
+                'position': 1,
+                'packet_hits': 200,
+                'byte_hits': 83883,
+                'id': '230',
+                'enabled': True,
+                'schedule': 'Always',
+                'log': 'all',
+                'l3_src': 'any',
+                'l3_dst': 'any',
+                'service': 'HTTP',
+                'src_zone': 'port2',
+                'dst_zone': 'port3',
+                'action': 'Permit'
+            }]
+        }
+        """
+        raise NotImplementedError
+
+    def compliance_report(self, validation_file='validate.yml'):
+        """
+        Return a compliance report.
+
+        Verify that the device complies with the given validation file and writes a compliance
+        report file. See https://napalm.readthedocs.io/en/latest/validate.html.
+        """
+        return validate.compliance_report(self, validation_file=validation_file)
